﻿/**
 * @file
 * VuoCompiler interface.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This interface description may be modified and distributed under the terms of the GNU Lesser General Public License (LGPL) version 2 or later.
 * For more information, see http://vuo.org/license.
 */

#ifndef VUOCOMPILER_H
#define VUOCOMPILER_H

class VuoCompilerNodeClass;
#include "VuoCompilerInputEventPortClass.hh"
#include "VuoCompilerOutputEventPortClass.hh"
#include "VuoCompilerGeneratorPortClass.hh"
#include "VuoCompilerNode.hh"
#include "VuoCompilerGraph.hh"
#include "VuoCompilerType.hh"
#include "VuoRunner.hh"


/**
 * A compiler/linker for node classes and graphs.
 *
 * Once you've compiled and linked a graph, you can use VuoRunner to execute and communicate with it.
 */
class VuoCompiler
{
private:
	map<string, VuoCompilerNodeClass *> nodeClasses;
	map<string, VuoCompilerType *> types;
	vector<string> headerSearchPaths;
	vector<string> librarySearchPaths;
	vector<string> preferredLibrarySearchPaths;
	vector<string> frameworkSearchPaths;
	vector<string> preferredFrameworkSearchPaths;
	llvm::sys::Path clangPath;
	string telemetry;
	string target;
	bool isClangVerbose;

	void loadNodesInDirectory(string path);
	void getDependenciesForGraph(Linker &linker, Linker &preferredLibraryLocator, bool doAddMainFunction, vector<Module *> &moduleDependencies, vector<sys::Path> &libraryDependencies, vector<string> &frameworks);
	void getDependenciesForNodeClass(Linker &linker, Linker &preferredLibraryLocator, VuoCompilerNodeClass *nodeClass, vector<Module *> &moduleDependencies, vector<sys::Path> &libraryDependencies, vector<string> &frameworks);
	void getDependenciesForType(Linker &linker, Linker &preferredLibraryLocator, VuoCompilerType *type, vector<Module *> &moduleDependencies, vector<sys::Path> &libraryDependencies, vector<string> &frameworks);
	Module * linkGraphIntoModule(string inputPath, bool doLoadDynamicLibraries, bool doAddMainFunction, vector<string> &libraries, vector<string> &frameworks);
	void linkGraphIntoExecutableOrDynamicLibrary(string inputPath, string outputPath, bool isDylib);
	void linkModule(Module *&module, bool doLinkNativeDependencies, bool doAddMainFunction, vector<string> &libraries, vector<string> &frameworks);
	llvm::sys::Path findLibAtPreferredPath(string dependencyFile, Linker &linker, Linker &preferredLibraryLocator);
	void linkInModuleDependencies(Linker &linker, const vector<Module *> &moduleDependencies);
	void linkInLibraryDependencies(Linker &linker, const vector<sys::Path> &libraryDependencies, bool doLoadDynamicLibraries, vector<string> &libraries);
	void linkBitcodeFile(string inputPath, string outputPath, const vector<string> &libraries, const vector<string> &frameworks, bool isDylib);
	Module * readModuleFromC(string inputPath);
	static Module * readModuleFromBitcode(string inputPath);
	static bool writeModuleToBitcode(Module *module, string outputPath);
	void setTargetForModule(Module *module, string target = "");
	static vector<string> getCoreVuoDependencies(bool doAddMainFunction);
	llvm::sys::Path getVuoFrameworkPath();
	static string getModuleNameForPath(string path);
	llvm::sys::Path getClangPath(void);
	static string getTypeInfo(Type *type);

	friend class TestVuoCompiler;
	friend class TestVuoCompilerModule;
	friend class TestVuoCompilerGraph;

public:
	VuoCompiler(void);
	void addModuleSearchPath(string path);
	void realizePortTypes(void);
	void compileGraph(VuoCompilerGraph *graph, string outputPath);
	void compileGraph(string inputPath, string outputPath);
	void compileGraphString(const string &graphString, string outputPath);
	void compileNodeClass(string inputPath, string outputPath);
	void linkGraphIntoExecutable(string inputPath, string outputPath);
	void linkGraphIntoDynamicLibrary(string inputPath, string outputPath);
	VuoCompilerNodeClass * getNodeClass(const string &id) const;
	map<string, VuoCompilerNodeClass *> getNodeClasses(void);
	VuoCompilerType * getType(const string &id) const;
	void listNodeClasses(const string &format = "") const;
	static map<string, string> getPortDeclarationInfo(VuoCompilerInputEventPortClass *portClass);
	static map<string, string> getPortDeclarationInfo(VuoCompilerOutputEventPortClass *portClass);
	static map<string, string> getPortDeclarationInfo(VuoCompilerGeneratorPortClass *portClass);
	void addHeaderSearchPath(const string &path);
	void addLibrarySearchPath(const string &path);
	void addPreferredLibrarySearchPath(const string &path);
	void clearPreferredLibrarySearchPaths();
	void addFrameworkSearchPath(const string &path);
	void addPreferredFrameworkSearchPath(const string &path);
	void setTelemetry(const string &telemetry);
	void setTarget(const string &target);
	void setClangVerbose(bool isClangVerbose);
	void setClangPath(const string &clangPath);

	static VuoRunner * newSeparateProcessRunnerFromCompositionFile(string compositionFilePath);
	static VuoRunner * newSeparateProcessRunnerFromCompositionString(string composition);
	static VuoRunner * newCurrentProcessRunnerFromCompositionFile(string compositionFilePath);
	static VuoRunner * newCurrentProcessRunnerFromCompositionString(string composition);
};


#endif
