/**
 * @file
 * VuoRunner interface.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This interface description may be modified and distributed under the terms of the GNU Lesser General Public License (LGPL) version 2 or later.
 * For more information, see http://vuo.org/license.
 */

#ifndef VUORUNNER_H
#define VUORUNNER_H

#include <dispatch/dispatch.h>
#include <stdexcept>
#include "VuoTelemetry.h"
#include "VuoPublishedPort.hh"
class VuoRunnerDelegate;


/**
 * This class runs a graph and manages communication with a graph produced by @c VuoCompiler.
 *
 * It can run a graph in the current process using the @c newCurrentProcessRunnerFromDynamicLibrary(string dylibPath) factory method,
 * or in a separate process using the @c newSeparateProcessRunnerFromExecutable(string executablePath) factory method.
 *
 * See also the following VuoCompiler methods which compile a graph and return a VuoRunner:
 *
 *    - @ref VuoCompiler::newSeparateProcessRunnerFromCompositionFile
 *    - @ref VuoCompiler::newSeparateProcessRunnerFromCompositionString
 *    - @ref VuoCompiler::newCurrentProcessRunnerFromCompositionFile
 *    - @ref VuoCompiler::newCurrentProcessRunnerFromCompositionString
 *
 * VuoRunner allows a client to control the graph (start, change input port values, generate events, hibernate, stop)
 * and receive data from the graph (output port values, uptime, node executions, edge hits).
 *
 * @c start(), @c startPaused(), and any member functions that send control requests to the graph are
 * not mutually thread-safe. They must not be executed concurrently.
 *
 * If @c start() or @c startPaused() is called, then @c stop() must also be called.
 */
class VuoRunner
{
private:
	string executablePath;  ///< The path to the linked graph executable.
	string dylibPath;  ///< The path to the linked graph dynamic library.
	void *dylibHandle;  ///< A handle to the linked graph dynamic library.
	bool stopped;	///< True if the graph is in a stopped state (either never started or started then stopped).
	bool listenCanceled;  ///< True if the @c listen() loop should end.
	dispatch_semaphore_t stoppedSemaphore;  ///< Signaled when the composition stops.
	dispatch_semaphore_t listenSemaphore;  ///< Signaled when the @c listen() loop ends.

	void *ZMQContext;	///< The context used to initialize sockets.
	void *ZMQControl;	///< The control socket. Not thread-safe.
	void *ZMQTelemetry;	///< The telemetry socket. Not thread-safe.
	string ZMQTelemetryURL; ///< The URL for the telemetry socket.

	VuoRunnerDelegate *delegate;  ///< The delegate that receives telemetry data. May be null.
	dispatch_queue_t delegateQueue;  ///< Synchronizes access to @c delegate.

	vector<VuoPublishedPort *> publishedInputPorts;
	vector<VuoPublishedPort *> publishedOutputPorts;
	map<string, VuoPublishedPort *> publishedOutputPortForIdentifier;  ///< Maps private port identifiers to the published output ports they're connected to.
	bool arePublishedInputPortsCached;  ///< True if the list of published input ports has been retrieved and cached.
	bool arePublishedOutputPortsCached; ///< True if the list of published output ports has been retrieved and cached.

	dispatch_semaphore_t anyPublishedPortEventSemaphore;  ///< Semaphore for waiting on an event from any published port.
	bool anyPublishedPortEventSignaled;  ///< Helps with @c anyPublishedPortEventSemaphore.
	void saturating_semaphore_signal(dispatch_semaphore_t dsema, bool *sent);
	void saturating_semaphore_wait(dispatch_semaphore_t dsema, bool *sent);

	VuoRunner(void);
	void start(bool isPaused);
	void listen(void);
	void vuoControlRequestSend(enum VuoControlRequest request, zmq_msg_t *messages, unsigned int messageCount);
	void vuoControlReplyReceive(enum VuoControlReply expectedReply);
	vector<string> receiveListOfStrings(void *socket);
	vector<VuoPublishedPort *> getCachedPublishedPorts(bool input);
	vector<VuoPublishedPort *> refreshPublishedPorts(bool input);
	bool isInCurrentProcess(void);

public:
	static VuoRunner * newSeparateProcessRunnerFromExecutable(string executablePath);
	static VuoRunner * newCurrentProcessRunnerFromDynamicLibrary(string dylibPath);
	~VuoRunner(void);
	void start(void);
	void startPaused(void);
	void runOnMainThread(void);
	void drainMainDispatchQueue(void);
	void pause(void);
	void unpause(void);
	void hibernate(void);
	void stop(void);
	void waitUntilStopped(void);
	void setInputPortValue(string portIdentifier, string valueAsString);
	string getOutputPortValue(string portIdentifier);
	void setPublishedInputPortValue(VuoPublishedPort *port, string valueAsString);
	void generatePublishedInputPortEvent(VuoPublishedPort *port);
	void waitForAnyPublishedOutputPortEvent(void);
	string getPublishedOutputPortValue(VuoPublishedPort *port);
	vector<VuoPublishedPort *> getPublishedInputPorts(void);
	vector<VuoPublishedPort *> getPublishedOutputPorts(void);
	VuoPublishedPort * getPublishedInputPortWithName(string name);
	VuoPublishedPort * getPublishedOutputPortWithName(string name);
	string getExecutablePath(void);
	bool isStopped(void);
	void setDelegate(VuoRunnerDelegate *delegate);
};


/**
 * An abstract class to be implemented by a client that will receive telemetry data from the graph.
 *
 * Inherit from this class if you want to implement all of the delegate methods.
 * If you'd only like to implement some delegate methods, inherit from @c VuoRunnerDelegateAdapter.
 */
class VuoRunnerDelegate
{
public:
	/**
	 * This delegate method is invoked twice per second, to provide a heartbeat indicating that the composition is still alive.
	 * It also provides some basic usage stats.
	 * @param utime The number of microseconds this process has spent in user-mode execution.
	 * @param stime The number of microseconds spent in the system executing on behalf of this process.
	 */
	virtual void receivedTelemetryStats(unsigned long utime, unsigned long stime) = 0;

	/**
	 * This delegate method is invoked every time a node has started executing.
	 * @param nodeIdentifier A unique identifier representing the node that started executing.
	 * @todo What is the format of the identifier?  How does it match up with the graphviz source?
	 */
	virtual void receivedTelemetryNodeExecutionStarted(string nodeIdentifier) = 0;

	/**
	 * This delegate method is invoked every time a node has finished executing.
	 * @param nodeIdentifier A unique identifier representing the node that finished executing.
	 * @todo What is the format of the identifier?  How does it match up with the graphviz source?
	 */
	virtual void receivedTelemetryNodeExecutionFinished(string nodeIdentifier) = 0;

	/**
	 * This delegate method is invoked every time an event has traversed a cable.
	 * @param edgeIdentifier A unique identifier representing the cable that was traversed.
	 * @todo What is the format of the identifier?  How does it match up with the graphviz source?
	 */
	virtual void receivedTelemetryEdgeHit(string edgeIdentifier) = 0;

	/**
	 * This delegate method is invoked every time any output port value has changed.
	 * @param portIdentifier A unique identifier representing the port whose value has changed.
	 * @param valueAsString The new value of the port, serialized.
	 * @todo What is the format of the identifier?  How does it match up with the graphviz source?
	 */
	virtual void receivedTelemetryPortValueChanged(string portIdentifier, string valueAsString) = 0;

	/**
	 * This delegate method is invoked every time any published output port value has changed.
	 * @param port The @c VuoPublishedPort whose value has changed (see @c VuoRunner::getPublishedOutputPorts() and @c VuoRunner::getPublishedOutputPortWithName()).
	 * @param valueAsString The new value of the port, serialized.
	 */
	virtual void receivedTelemetryPublishedPortValueChanged(VuoPublishedPort *port, string valueAsString) = 0;

	virtual ~VuoRunnerDelegate() = 0;  // Fixes "virtual functions but non-virtual destructor" warning
};

#define VUO_UNUSED_VARIABLE __attribute__((unused))

/**
 * A class provided for convenience when inheriting from VuoRunnerDelegate. If inheriting from this
 * class instead of VuoRunnerDelegate directly, the derived class does not need to implement all
 * VuoRunnerDelegate methods.
 */
class VuoRunnerDelegateAdapter : public VuoRunnerDelegate
{
	virtual void receivedTelemetryStats(unsigned long VUO_UNUSED_VARIABLE utime, unsigned long VUO_UNUSED_VARIABLE stime) { }
	virtual void receivedTelemetryNodeExecutionStarted(string VUO_UNUSED_VARIABLE nodeIdentifier) { }
	virtual void receivedTelemetryNodeExecutionFinished(string VUO_UNUSED_VARIABLE nodeIdentifier) { }
	virtual void receivedTelemetryEdgeHit(string VUO_UNUSED_VARIABLE edgeIdentifier) { }
	virtual void receivedTelemetryPortValueChanged(string VUO_UNUSED_VARIABLE portIdentifier, string VUO_UNUSED_VARIABLE valueAsString) { }
	virtual void receivedTelemetryPublishedPortValueChanged(VuoPublishedPort VUO_UNUSED_VARIABLE *port, string VUO_UNUSED_VARIABLE valueAsString) { }
};

#endif
